/* Copyright (C) 2009, Alexander Graf <agraf@znc.in>
 * No rights reserved. */

.text
.arm

.global _start
_start:
	ldr	pc, =bootstrap
	ldr	pc, =undef
	ldr	pc, =undef
	ldr	pc, =undef
	ldr	pc, =abort
	nop
	ldr	pc, =undef
	ldr	pc, =undef

/* called on data abort exception (may occur on RAM test if RAM is unmapped) */
abort:
	ldr	r4,	=data_abort
	bl	uart_putstring
1:	b	1b

/* called on every exception which is neither caused by data abort nor by an
 * reset. */
undef:
	ldr	r4,	=exc
	bl	uart_putstring
1:	b	1b

/* called on reset exception */
bootstrap:
	ldr	r0,	=0xe01fc000	/* base of some system control regs */

	/* enable main oscillator */
	ldr	r1,	=(1 << 5)	/* main oscillator enable */
	str	r1,	[r0, #0x1a0]
1:	ldr	r1,	[r0, #0x1a0]	/* wait for oscillator to be ready */
	ands	r1, 	r1,	#(1 << 6)
	beq	1b

	/* initalize pll */
	ldr	r2,	=0xaa		/* feed sequence */
	ldr	r3,	=0x55
	ldr	r1,	=0		/* disable PLL */
	str	r1,	[r0, #0x080]
	str	r2,	[r0, #0x08c]
	str	r3,	[r0, #0x08c]
	ldr	r1,	=1		/* use main osci as PLL source */
	str	r1,	[r0, #0x10c]
	ldr	r1,	=(23 | 1 << 16)	/* M: 24, D: 1 */
	str	r1,	[r0, #0x084]
	str	r2,	[r0, #0x08c]
	str	r3,	[r0, #0x08c]
	ldr	r1,	=1		/* enable */
	str	r1,	[r0, #0x080]
	str	r2,	[r0, #0x08c]
	str	r3,	[r0, #0x08c]
	ldr	r1,	=3		/* CCLKCFG, divisor = 4 */
	str	r1,	[r0, #0x104]
	ldr	r1,	=5		/* USBCLKCFG, divisor = 6 */
	str	r1,	[r0, #0x108]
1:	ldr	r1,	[r0, #0x088]	/* wait for pll to be locked */
	ands	r1,	r1,	#(1 << 26)
	beq	1b
	ldr	r1,	=3		/* enable & connect pll */
	str	r1,	[r0, #0x080]
	str	r2,	[r0, #0x08c]
	str	r3,	[r0, #0x08c]

	/* set CCLK/1 as peripherial clock */
	ldr	r1,	=0x55515555
	str	r1,	[r0, #0x1a8]
	ldr	r1,	=0x11555455
	str	r1,	[r0, #0x1ac]

	/* initalize MAM */
	ldr	r1,	=4		/* 4 clocks as MAM timing */
	str	r1,	[r0, #0x004]
	ldr	r1,	=2		/* fully enable */
	str	r1,	[r0, #0x000]

	/* configure UART0 pins as UART0 pins */
	ldr	r0,	=0xe002c000
	ldr	r1,	=(1 << 4 | 1 << 6)
	str	r1,	[r0, #0x0]

	/* initalize UART0 */
	ldr	r3,	=0xe000c000	/* uart base */
	ldr	r1,	=7		/* enable and reset FIFO */
	str	r1,	[r3, #0x8]
	ldr	r1,	=(1 << 7)	/* set divisor latch access bit */
	str	r1,	[r3, #0xc]
	ldr	r1,	=((72000000 / 115200 / 16) % 256) /* set divisor */
	str	r1,	[r3, #0x0]
	ldr	r1,	=((72000000 / 115200 / 16) / 256)
	str	r1,	[r3, #0x4]
	ldr	r1,	=3		/* 8 bits per byte */
	str	r1,	[r3, #0xc]

	/* from now on, r3, r4 and r5 are occupied by UART driver.
	 * r3: all time (keeps UART base address), r4 and r5 are altered when
	 * uart_putstring is called */

	ldr	r4,	=uart_works
	bl	uart_putstring

	/* SDRAM initalization */
	/* configure PINs */
	ldr	r0,	=0xe002c000	/* pinsel base */
	ldr	r1,	=0x50000000
	str	r1,	[r0, #0x10]	/* CS2, CS3 */
	ldr	r1,	=0x55015555
	str	r1,	[r0, #0x14]	/* CAS, RAS, CLKOUT0, CLKOUT1, DYCS0..DYCS3, CKEOUT0, DQMOUT0..DQMOUT3 */
	ldr	r1,	=0x55555555
	str	r1,	[r0, #0x18]	/* D0..D15 */
	str	r1,	[r0, #0x1c]	/* D16..D31 */
	str	r1,	[r0, #0x20]	/* A0..A15 */
	str	r1,	[r0, #0x24]	/* A16..A23, OE, WE, BLS0..BLS3, CS0, CS1 */
	/* now disable resistors for these pins */
	ldr	r1,	=0xa0000000
	str	r1,	[r0, #0x50]
	ldr	r1,	=0xaa02aaaa
	str	r1,	[r0, #0x54]
	ldr	r1,	=0xaaaaaaaa
	str	r1,	[r0, #0x58]
	str	r1,	[r0, #0x5c]
	str	r1,	[r0, #0x60]
	str	r1,	[r0, #0x64]
	/* now configure EMC */
	ldr	r0,	=0xffe08000	/* EMC base */
	ldr	r1,	=1
	str	r1,	[r0, #0x000]	/* EMC enable */
	ldr	r1,	=1
	str	r1,	[r0, #0x028]	/* set read strategy */
	ldr	r1,	=(3 | 3 << 8)
	str	r1,	[r0, #0x104]	/* RAS & CAS timings */
	ldr	r1,	=4
	str	r1,	[r0, #0x030]	/* tRP timing */
	ldr	r1,	=8
	str	r1,	[r0, #0x034]	/* tRAS timing */
	ldr	r1,	=5
	str	r1,	[r0, #0x038]	/* tSREX timing */
	ldr	r1,	=2
	str	r1,	[r0, #0x03c]	/* tAPR timing */
	ldr	r1,	=5
	str	r1,	[r0, #0x040]	/* tDAL timing */
	ldr	r1,	=3
	str	r1,	[r0, #0x044]	/* tWR timing */
	ldr	r1,	=11
	str	r1,	[r0, #0x048]	/* tRC timing */
	ldr	r1,	=11
	str	r1,	[r0, #0x04c]	/* tRFC timing */
	ldr	r1,	=5
	str	r1,	[r0, #0x050]	/* tXSR timing */
	ldr	r1,	=3
	str	r1,	[r0, #0x054]	/* tRRD timing */
	ldr	r1,	=8
	str	r1,	[r0, #0x058]	/* tMRD timing */
	ldr	r1,	=(2 << 7 | 1 << 9 | 1 << 14)
	str	r1,	[r0, #0x100]	/* address mappings */
	/* init sequence */
	ldr	r1,	=0
1:	add	r1,	r1,	#1
	cmp	r1,	#128
	bne	1b
	ldr	r1,	=0x183
	str	r1,	[r0, #0x020]
	ldr	r1,	=0
1:	add	r1,	r1,	#1
	cmp	r1,	#128
	bne	1b
	ldr	r1,	=0x101
	str	r1,	[r0, #0x020]
	ldr	r1,	=1
	str	r1,	[r0, #0x024]
	ldr	r1,	=0
1:	add	r1,	r1,	#1
	cmp	r1,	#128
	bne	1b
	ldr	r1,	=19
	str	r1,	[r0, #0x024]
	ldr	r1,	=0x081
	str	r1,	[r0, #0x020]
	ldr	r1,	=(0xa0000000 | (0x32 << 12))
	ldr	r2,	[r1]
	ldr	r1,	=0x000
	str	r1,	[r0, #0x020]
	ldr	r1,	[r0, #0x100]
	orr	r1,	r1,	#(1 << 19)
	str	r1,	[r0, #0x100]

	ldr	r4,	=sdram_init
	bl	uart_putstring

	/* do test of RAM */
	ldr	r4,	=ramtest_int
	bl	uart_putstring
	ldr	r0,	=0x40000000
	ldr	r1,	=(64 << 10)
	bl	ramtest
	ldr	r4,	=ramtest_ext
	bl	uart_putstring
	ldr	r0,	=0xa0000000
	ldr	r1,	=(16 << 20)
	bl	ramtest

	ldr	r0,	=0x3fffc000	/* fast gpio base */
	
	/* set led gpio pin direction */
	ldr	r1,	=(1 << 25 | 1 << 26 | 1 << 27)
	str	r1,	[r0, #0x40]

	/* clear green and red LED */
	ldr	r1,	=(1 << 26 | 1 << 27)
	str	r1,	[r0, #0x5c]

	/* set blue LED */
	ldr	r1,	=(1 << 25)
	str	r1,	[r0, #0x58]

	ldr	r4,	=led_flashes
	bl	uart_putstring

	/* let green+red LED flash */
	ldr	r1,	=(1 << 26 | 1 << 27)
1:	ldr	r2,	=0
2:	add	r2,	r2,	#1
	cmp	r2,	#(1 << 25)
	bne	2b
	str	r1,	[r0, #0x58]
	ldr	r2,	=0
2:	add	r2,	r2,	#1
	cmp	r2,	#(1 << 16)
	bne	2b
	str	r1,	[r0, #0x5c]
	b	1b

/* prints \0-terminated string on UART
 * expects r3 to contain uart base and r4 to contain pointer to string.
 * modificates r4 and r5 */
uart_putstring:
1:	ldrb	r5,	[r4]
	cmp	r5,	#0
	bxeq	r14
	str	r5,	[r3]
	add	r4,	r4,	#1
2:	ldr	r5,	[r3, #0x14]
	ands	r5,	r5,	#(1 << 5)
	beq	2b
	b	1b

/* tests a piece of RAM by writing and reading the address.
 * expects r0 to contain ram base and r1 to contain ram size
 * modificates r2, r4, r5 */
ramtest:
	ldr	r2,	=0		/* counter */
1:	str	r2,	[r0, r2]
	ldr	r4,	[r0, r2]
	cmp	r2,	r4
	bne	1f
	add	r2,	r2,	#4
	cmp	r2,	r1
	bne	1b
1:	ldreq	r4,	=ramtest_succeed
	ldrne	r4,	=ramtest_failed
	b	uart_putstring		/* uart_putstring ends with bx r14, so
					 * this does return to our return link
					 */

.section .rodata
data_abort:	.asciz "Data Abort (attempted to access unmapped memory).\n"
exc:		.asciz "Unexpected exception happened.\n"
uart_works:	.asciz "UART seems to work (if you can read this).\n"
sdram_init:	.asciz "SDRAM initalized.\n"
ramtest_int:	.asciz "Testing internal RAM: "
ramtest_ext:	.asciz "Testing external RAM: "
ramtest_succeed:.asciz "Works.\n"
ramtest_failed:	.asciz "Failed.\n"
led_flashes:	.asciz "Now blue LED is on and red+green LED flashes.\n"
